const HTTPMethods = [
    'GET',
    'POST',
    'PUT',
    'DELETE',
    'PATCH',
    'HEAD',
    'TRACE',
    'OPTIONS',
    'CONNECT',
];
const isHTTPMethod = (s) => HTTPMethods.includes(s);
// eslint-disable-next-line @typescript-eslint/no-explicit-any
const isBlob = (e) => e != null && e instanceof Blob;
const checkHTTPMethod = (s) => {
    if (!isHTTPMethod(s)) {
        throw new Error('Not HTTP Method');
    }
    return s;
};
// eslint-disable-next-line @typescript-eslint/no-explicit-any
const checkBlob = (e) => {
    if (!isBlob(e)) {
        throw new Error('Not a Blob');
    }
    return e;
};
const buildGMHeaders = (headers) => {
    return Object.fromEntries(headers.entries());
};
/**
 * Encode binary data as isomorphic-encoded string.
 *
 * Binary data cannot be encoded as JavaScript string directly
 * because some binary data may contain some binary sequences
 * which are treated specially in the text encoding.
 * GM.xmlhttpRequest requires that each single byte (uint8) is encoded as CharCode.
 * By doing this, it is possible for GM.xmlHttpRequest to obtain the original binary data.
 *
 * https://developer.mozilla.org/en-US/docs/Web/API/btoa
 * https://infra.spec.whatwg.org/#isomorphic-encode
 * https://infra.spec.whatwg.org/#isomorphic-decode
 *
 * GreaseMonkey
 *   https://github.com/greasemonkey/greasemonkey/blob/efd22a93121225ada47f2fe9f021af0ab6100c21/src/bg/on-user-script-xhr.js#L140-L146
 */
const encodeArrayBufferAsIsomorphicEncodedString = (b) => {
    const view = new Uint8Array(b);
    const batch = 1024;
    const result = [];
    for (let i = 0; i < view.length; i += batch) {
        const str = String.fromCharCode(...view.subarray(i, i + batch));
        result.push(str);
    }
    return result.join('');
};
const buildData = async (request) => {
    const arrayBuffer = await request.arrayBuffer();
    const result = encodeArrayBufferAsIsomorphicEncodedString(arrayBuffer);
    if (result.length === 0)
        return undefined;
    return result;
};
const buildWithCredentials = (request, location) => {
    const url = new URL(request.url);
    switch (request.credentials) {
        case 'include':
            return true;
        case 'same-origin':
            return location.origin === url.origin;
        case 'omit':
            return false;
        default:
            throw new TypeError('Unknown credentials type.');
    }
};
const buildDetails$1 = async (request) => {
    if (request.redirect !== 'follow') {
        throw new TypeError('Only follow is supported');
    }
    const withCredentials = buildWithCredentials(request, globalThis.location);
    const headers = buildGMHeaders(request.headers);
    return {
        method: checkHTTPMethod(request.method),
        url: request.url,
        headers,
        binary: true,
        timeout: undefined,
        responseType: 'blob',
        // Compatibility
        data: await buildData(request),
        // Tampermonkey
        anonymous: !withCredentials,
        // GreaseMonkey
        withCredentials,
    };
};
/**
 * Parse response headers string and covert it to Headers object.
 *
 * https://developer.mozilla.org/ja/docs/Web/API/XMLHttpRequest/getAllResponseHeaders
 */
const parseXHRHeaders = (gmHeaders) => {
    const headers = new Headers();
    if (gmHeaders == null) {
        return headers;
    }
    const trimmedHeaders = gmHeaders.trim();
    if (trimmedHeaders.length === 0) {
        return headers;
    }
    let name = '';
    let value = '';
    // A sequence '\r\n' is strictly required.
    trimmedHeaders.split(/(?:\r\n)+/).forEach((line) => {
        const index = line.indexOf(':');
        // The line with no colon (:) will be ignored.
        if (index < 0) {
            // obs-fold
            if (/^[ \t]/.test(line) && name.length > 0 && value.length > 0) {
                const appendValue = line.trim();
                headers.set(name, `${value}, ${appendValue}`);
            }
            return;
        }
        name = line.substring(0, index);
        // For Firefox-styled multi-lined headers.
        [value] = line
            .substring(index + 1)
            .trim()
            .split(/[\r\n]+/);
        // Check that only allowed character is used. Ignored if invalid.
        //
        // Limitation for Set-Cookie:
        //   According to 3.2.2 [RFC7320], Set-Cookie can appear in header multiple times.
        //   However, only the first one is available with Chrome XHR.
        //   On the other hand, Firefox generates a LF(\n)-separated string for multiple Set-Cookie(s)
        //   --- it's considered as incompatible with specification because LF is not allowed ---
        //   but Headers () does not support multiple values.
        //   As for other headers, multiple headers can be comma-separated text. However, Set-Cookie cannot be.
        if (!/^[ \t\x21-\x7e\x80-\xff]*$/.test(value))
            return;
        headers.set(name, value);
    });
    return headers;
};
const algHash = {
    sha256: 'SHA-256',
    sha384: 'SHA-384',
    sha512: 'SHA-512',
};
/**
 * https://w3c.github.io/webappsec-subresource-integrity/#the-integrity-attribute
 */
const verifyIntegrityWith = (crypto) => async (integrity, data) => {
    if (crypto?.subtle?.digest == null) {
        throw new Error('digest function is not available.');
    }
    const match = integrity.match(/^(sha(?:256|384|512))-([a-zA-Z0-9+/=]+)$/);
    if (!match) {
        throw new TypeError('unsupported alg or integrity format');
    }
    const [, alg, expected] = match;
    const hashAlgorithm = algHash[alg] || alg;
    const digest = await globalThis.crypto.subtle.digest(hashAlgorithm, await data.arrayBuffer());
    const actual = btoa(encodeArrayBufferAsIsomorphicEncodedString(digest));
    if (actual !== expected) {
        throw new Error('Integrity verification failed');
    }
};
const verifyIntegrity = verifyIntegrityWith(globalThis.crypto);
/**
 * Convert GM.Response to Response
 */
// eslint-disable-next-line @typescript-eslint/no-explicit-any
const buildResponse = (res) => {
    const body = checkBlob(res.response);
    return new Response(body, {
        status: res.status,
        statusText: res.statusText,
        headers: parseXHRHeaders(res.responseHeaders),
    });
};
/**
 * Supported status:
 *   init:
 *     - mode:           Ignored.
 *     - credentials:    Supported.
 *     - cache:          Ignored.
 *     - redirect:       Ignored. (if non-"follow" value is specified, error log will be printed.)
 *     - integrity:      Partially supported. Multiple integrities are not supported.
 *     - signal:         Supported.
 *     - referrer:       Ignored. TBD.
 *     - referrerPolicy: Ignored
 *   response object:
 *     - redirect: always false even if redirect.
 *     - type:     always "default"
 *
 * GreaseMonkey: https://wiki.greasespot.net/GM.xmlHttpRequest
 *   https://github.com/greasemonkey/greasemonkey/blob/efd22a93121225ada47f2fe9f021af0ab6100c21/src/bg/api-provider-source.js#L252
 *   https://github.com/greasemonkey/greasemonkey/blob/efd22a93121225ada47f2fe9f021af0ab6100c21/src/bg/on-user-script-xhr.js#L23
 *
 * Tampermonkey https://www.tampermonkey.net/documentation.php#GM_xmlhttpRequest
 *   NOTE: very old version of Tampermonkey
 *   https://github.com/Tampermonkey/tampermonkey/blob/07f668cd1cabb2939220045839dec4d95d2db0c8/src/xmlhttprequest.js
 *
 * Violentmonkey: https://violentmonkey.github.io/api/gm/#gm_xmlhttprequest
 *
 * Fetch Standard https://fetch.spec.whatwg.org/#fetch-method
 */
const gmFetch = async (resource, init = {}) => {
    if (GM.xmlHttpRequest == null) {
        throw new TypeError('GM.xmlHttpRequest is not listed in @grant.');
    }
    const request = new Request(resource, init);
    const details = await buildDetails$1(request);
    if (request.signal.aborted) {
        throw new DOMException('The request was aborted.', 'AbortError');
    }
    return new Promise((resolve, reject) => {
        const control = GM.xmlHttpRequest({
            ...details,
            onload(res) {
                const checkAndBuildResponse = async () => {
                    if (request.integrity) {
                        const body = checkBlob(res.response);
                        await verifyIntegrity(request.integrity, body);
                    }
                    return buildResponse(res);
                };
                checkAndBuildResponse()
                    .then((response) => resolve(response))
                    .catch((err) => reject(err));
            },
            onerror() {
                reject(new TypeError('NetworkError'));
            },
            ontimeout() {
                reject(new TypeError('TimeoutError'));
            },
            onabort() {
                reject(new DOMException('The request was aborted.', 'AbortError'));
            },
        });
        if (request.signal !== undefined) {
            if (control != null) {
                request.signal.addEventListener('abort', () => {
                    control.abort();
                });
            }
            else {
                reject(new TypeError('abort is not supported in this implementation.'));
            }
        }
    });
};

const buildDetails = (args) => {
    if (typeof args[0] === 'string') {
        const url = args[0].toString();
        const name = args[1];
        if (name == null) {
            throw new Error('name is empty.');
        }
        return { url, name };
    }
    return args[0];
};
/**
 * GM_download wrapper.
 */
const gmDownload = (...args) => {
    if (typeof GM_download === 'undefined') {
        throw new Error('GM_download is not listed in @grant or not supported on this environment.');
    }
    const { signal, ...details } = buildDetails(args);
    if (signal?.aborted) {
        throw new DOMException('The request was aborted.', 'AbortError');
    }
    return new Promise((resolve, reject) => {
        const control = GM_download({
            ...details,
            onload() {
                resolve();
            },
            onerror(err) {
                const errDetails = err.details != null ? `: ${err.details}` : '';
                reject(new TypeError(`DownloadError: ${err.error}${errDetails}`));
            },
            ontimeout() {
                reject(new TypeError('TimeoutError'));
            },
        });
        if (signal != null) {
            signal.addEventListener('abort', () => control.abort());
        }
    });
};

export { gmDownload, gmFetch };
